"""PMeter a tool to measure the TCP/UDP network conditions that the running host experiences

Usage:
  pmeter_cli.py measure <INTERFACE> [-K=KER_BOOL -N=NET_BOOL -F=FILE_NAME -S=STD_OUT_BOOL --interval=INTERVAL --measure=MEASUREMENTS --length=LENGTH --user=USER --file_name=FILENAME --folder_path=FOLDERPATH --folder_name=FOLDERNAME]
  pmeter_cli.py carbon <IP> [--max_hops=<MAX_HOPS> --save_per_ip=<SAVE_PER_IP> --save_time=<TIME>]

Commands:
    measure     The command to start measuring the computers network activity on the specified network devices. This command accepts a list of interfaces that you wish to monitor
    carbon      Measure the carbon footprint from the cli's host to the destination ip

Options:
  -h --help                Show this screen
  --version                Show version
  -N --measure_network     Set if we monitor only the network interface [default: True]
  -K --measure_kernel      Set if we monitor only the kernel [default: True]
  -U --measure_udp         Set UDP monitoring only [default: True]
  -T --measure_tcp         Set TCP monitoring only [default: True]
  -S --enable_std_out      Disable printing the results to standard output [default: False]
  --file_name=FILENAME     Set the file name used to measure [default: pmeter_measure.txt]
  --folder_path=FOLDERPATH Set the path to store the measurement files in the default is the users home directory under
  --folder_name=FOLDERNAME Set the folder name for pmeter to dump data into
  --interval=INTERVAL      Set the time to run the measurement in the format HH:MM:SS [default: 00:00:01]
  --measure=MEASUREMENTS   The max number of times to measure your system. Set this value to 0 to ignore this and use only the length [default: 1]
  --length=LENGTH          The amount of time to run for: 5w, 4d 3h, 2m, 1s are some examples of 5 weeks, 4 days, 3 hours, 2 min, 1 sec. Set this value to '-1s' to ignore this field and use only measurement [default: 10s]
  --user=USER              This will override the user we try to pick up from the environment variable(ODS_USER). If no user is passed then we will not submit the data generated to the ODS backend [default: ]
  --max_hops=MAX_HOPS      The maximum number of hops [default: 64]
  --save_per_ip=SAVE_PER_IP Save carbon intensity per IP address [default: False]
  --save_time=<TIME>        Capture timestamp when taking carbon footprint [default: False]
"""
import json
import math
from pathlib import Path

import pandas as pd
from docopt import docopt
from helpers import constants
from datetime import datetime, timedelta
from helpers.file_writer import ODS_Metrics
import time
import copy
import requests
from pandas import DataFrame, read_json
import os
from Tracert import tracert

# conf.use_pcap = True
old_measure_dict = {}  # this is hackish but the keyis the interface name and for every metric we run of that interface name we replace and then run


def convert_to_endate(length):
    end_date = datetime.now()
    num = int(length[:-1])
    if "s" in length:
        end_date += timedelta(seconds=num)
    if "m" in length:
        end_date += timedelta(minutes=num)
    if "d" in length:
        end_date += timedelta(days=num)
    if "w" in length:
        end_date += timedelta(weeks=num)
    if "h" in length:
        end_date += timedelta(hours=num)
    return end_date


# run measurement using time only
def measure_using_length(interface_list, metric, folder_path, file_name, folder_name, measure_tcp=True,
                         measure_udp=True, measure_kernel=True, measure_network=True, print_to_std_out=False,
                         latency_host="http://google.com", interval=1, length="0s"):
    end_date = convert_to_endate(length)
    current_date = datetime.now()
    while (current_date < end_date):
        print("Current date is less than end date=", current_date < end_date)
        metric.measure_latency_rtt(latency_host)
        for intr_name in interface_list:
            metric.measure(intr_name, measure_tcp, measure_udp, measure_kernel, measure_network, print_to_std_out,
                           latency_host)
            if intr_name in old_measure_dict:
                metric.do_deltas(old_measure_dict[intr_name])
            old_measure_dict[intr_name] = copy.deepcopy(metric)
            metric.to_file(file_name=file_name, folder_path=folder_path, folder_name=folder_name)
        current_date = datetime.now()
        time.sleep(interval)


def measure_using_measurements(interface_list, metric, folder_path, file_name, folder_name, measure_tcp=True,
                               measure_udp=True, measure_kernel=True, measure_network=True, print_to_std_out=False,
                               latency_host="http://google.com", interval=1, measurement=1):
    for i in range(0, measurement):
        metric.measure_latency_rtt(latency_host)
        print("measurement: ", i)
        for intr_name in interface_list:
            metric.measure(intr_name, measure_tcp, measure_udp, measure_kernel, measure_network, print_to_std_out,
                           latency_host)
            if intr_name in old_measure_dict:
                metric.do_deltas(old_measure_dict[intr_name])
            old_measure_dict[intr_name] = copy.deepcopy(metric)
            metric.to_file(file_name=file_name, folder_path=folder_path, folder_name=folder_name)
        time.sleep(interval)


def begin_measuring(user, folder_path, file_name, folder_name, interface='', measure_tcp=True, measure_udp=True,
                    measure_kernel=True, measure_network=True, print_to_std_out=False, interval=1,
                    latency_host="http://google.com", measurement=1, length="0s"):
    metric = ODS_Metrics()
    metric.set_user(user)
    interface_list = []
    if interface == "all":
        for inter_name in metric.get_system_interfaces():
            interface_list.append(inter_name)
    else:
        interface_list.append(interface)
    if measurement == 0:
        print("Measuring by using the duration specified with interval")
        measure_using_length(interface_list=interface_list, metric=metric, folder_path=folder_path, file_name=file_name,
                             folder_name=folder_name, measure_tcp=measure_tcp, measure_udp=measure_udp,
                             measure_kernel=measure_kernel, measure_network=measure_network,
                             print_to_std_out=print_to_std_out, latency_host=latency_host, interval=interval,
                             length=length)
    if length == '0':
        print("Measuring by using the number of measurments to perform with interval")
        measure_using_measurements(interface_list=interface_list, metric=metric, folder_path=folder_path,
                                   file_name=file_name, folder_name=folder_name, measure_tcp=measure_tcp,
                                   measure_udp=measure_udp, measure_kernel=measure_kernel,
                                   measure_network=measure_network, print_to_std_out=print_to_std_out,
                                   latency_host=latency_host, interval=interval, measurement=measurement)
    else:
        measurements_counter = 0
        end_date = convert_to_endate(length)
        current_date = datetime.now()
        while current_date < end_date and measurements_counter < measurement:
            print("Current date= ", str(current_date), "is less than end date=", str(end_date), " is =",
                  current_date < end_date)
            print("currentMeasurement is less than the max measurements", measurements_counter < measurement)
            for intr_name in interface_list:
                metric.measure(intr_name, measure_tcp, measure_udp, measure_kernel, measure_network, print_to_std_out,
                               latency_host)
                if intr_name in old_measure_dict:
                    metric.do_deltas(old_measure_dict[intr_name])
                old_measure_dict[intr_name] = copy.deepcopy(metric)
                metric.to_file(file_name=file_name, folder_path=folder_path, folder_name=folder_name)
            current_date = datetime.now()
            measurements_counter += 1
            time.sleep(interval)


def traceroute(destination, max_hops=30):
    ping_iterable = tracert(destination=destination, max_steps=max_hops)
    ip_list = []
    for ping in ping_iterable:
        ip_list.append(ping.ip)
    return ip_list



def geo_locate_ips(ip_list) -> pd.DataFrame:
    # access_key = os.getenv("GEO_LOCATE_ACCESS_KEY")
    payload = [{"query": ip} for ip in ip_list]
    url = "http://ip-api.com/batch"
    r = requests.post(url=url, json=payload)
    # return read_json(r.json())
    return DataFrame(r.json())


def compute_carbon_per_ip(ip_df, store_format=False, save_time=False):
    # ip_df.dropna(inplace=True)
    # ip_df.reset_index(drop=True, inplace=True)
    auth_token = os.getenv("ELECTRICITY_MAPS_AUTH_TOKEN")
    headers = {
        'auth-token': str(auth_token)
    }
    # Split the string into a list with one element
    resp_list = []
    carbon_ip_map = {}
    for idx, row in ip_df.iterrows():
        cur_lat = row['lat']
        cur_long = row['lon']
        if math.isnan(cur_lat) or math.isnan(cur_long):
            continue
        cur_ip = row['query']
        params = {'lon': cur_long, 'lat': cur_lat}
        resp = requests.get(url="https://api-access.electricitymaps.com/free-tier/carbon-intensity/latest",
                            params=params, headers=headers)
        carbon_data_json = resp.json()
        carbon_intensity = carbon_data_json['carbonIntensity']
        carbon_ip_map[cur_ip] = {"carbon_intensity": carbon_intensity, "lat": cur_lat, "lon": cur_long}
        print(f"Lat:{cur_lat} Lon:{cur_long} IP:{cur_ip} CarbonIntensity:{carbon_intensity}")
        resp_list.append(resp)
    carbon_intensity_path_total = 0
    for ip in carbon_ip_map:
        carbon_intensity_path_total += carbon_ip_map[ip]['carbon_intensity']
    avg_carbon_network_path = carbon_intensity_path_total / len(carbon_ip_map)
    print("Average Carbon cost for network path:  ", avg_carbon_network_path)
    carbon_ip_map['time'] = datetime.now().isoformat()
    if store_format == True:
        to_file(carbon_ip_map, file_name='carbon_ip_map.json')
    else:
        to_file(data={'avgCarbon': avg_carbon_network_path})
    return avg_carbon_network_path


def to_file(data, file_path=None, file_name="carbon_pmeter.txt"):
    if file_path is None:
        file_path = os.path.join(Path.home(), ".pmeter", file_name)
    print(file_path)
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    j = json.dumps(data)
    with open(file_path, "a+") as f:
        f.write(j + "\n")


def main():
    arguments = docopt(__doc__, version='PMeter 1.0')
    if arguments['measure']:
        interface = arguments['<INTERFACE>']
        file_name = arguments['--file_name']
        folder_path = arguments['--folder_path']
        folder_name = arguments['--folder_name']
        network_only = arguments['--measure_network']
        kernel_only = arguments['--measure_kernel']
        udp_only = arguments['--measure_udp']
        tcp_only = arguments['--measure_tcp']
        std_out_print = arguments['--enable_std_out']
        interval = arguments['--interval']
        pause_between_measure = constants.get_sec(interval)
        times_to_measure = arguments['--measure']
        lengthOfExperiment = arguments['--length']
        user = arguments['--user']
        if folder_name is None: folder_name = ".pmeter"
        print("Parameters recieved: \n", "interface=", interface, "measure network=", network_only, "measure kernel=",
              kernel_only, "interval between measurements=", pause_between_measure, "times to measure default is 1=",
              times_to_measure, 'file_name=', file_name, 'folder_name=', folder_name, 'folder_path=', folder_path)
        begin_measuring(folder_path=folder_path, folder_name=folder_name, file_name=file_name, user=user,
                        interface=interface, measure_tcp=tcp_only, measure_kernel=kernel_only,
                        measure_network=network_only, measure_udp=udp_only, print_to_std_out=std_out_print,
                        interval=pause_between_measure, length=lengthOfExperiment, measurement=int(times_to_measure))
        print("Done Measuring")
    elif arguments['carbon']:
        store_format = arguments['--save_per_ip']
        save_time = arguments['--save_time']
        ip_list = traceroute(arguments['<IP>'], int(arguments['--max_hops']))
        print(f"IP's to source {ip_list}")
        ip_df = geo_locate_ips(ip_list)
        compute_carbon_per_ip(ip_df, store_format=bool(store_format), save_time=bool(save_time))


if __name__ == '__main__':
    main()



class NetworkMetrics:
    BANDWIDTH="bandwith"
    PACKET_LOSS="packet_loss"
    LINK_CAPACTITY="link_capacity"
    ROUND_TRIP_TIME="round_trip_time"
    BANDWIDTH_DELAY_PRODUT="bandwidth_delay_product"

class KernelMetrics:
    ACTIVE_CORES="active_cores"
    CPU_FREQUENCY="cpu_frequency"
    ENERGY_CONSUMPTION="energy_consumption"
    CPU_ARCHITECTURE="cpu_architecture"
    NETWORK_CARD_USAGE="nic_usage"
    CPU_USAGE="cpu_usage"

#Found this function here https://www.admin-magazine.com/HPC/Articles/Process-Network-and-Disk-Metrics
def bytes2human(n):
       # From sample script for psutils
   """
   >>> bytes2human(10000)
   '9.8 K'
   >>> bytes2human(100001221)
   '95.4 M'
   """
   symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
   prefix = {}
   for i, s in enumerate(symbols):
      prefix[s] = 1 << (i + 1) * 10
   for s in reversed(symbols):
      if n >= prefix[s]:
         value = float(n) / prefix[s]
         return '%.2f %s' % (value, s)
   return '%.2f B' % (n)

#Found this here to convert HH:MM:SS to seconds https://stackoverflow.com/questions/6402812/how-to-convert-an-hmmss-time-string-to-seconds-in-python
def get_sec(time_str):
    print("Parsing time=",time_str, "to seconds")
    """Get seconds from time."""
    h, m, s = time_str.split(':')
    return int(h) * 3600 + int(m) * 60 + int(s)


from email.policy import default
import multiprocessing
from pathlib import Path
import json
import datetime
import psutil
import platform
from tcp_latency import measure_latency
import os
import requests
import statistics
from datetime import datetime


class ODS_Metrics():

    def __init__(self):
        # kernel metrics
        self.interface = ""
        self.ods_user = ""
        self.active_core_count = 0
        self.cpu_frequency_max = 0.0
        self.cpu_frequency_current = 0.0
        self.cpu_frequency_min = 0.0
        self.energy_consumed = 0.0
        self.cpu_arch = ""
        # network metrics
        self.rtt = 0.0
        self.bandwidth = 0.0
        self.bandwidth_delay_product = 0.0
        self.packet_loss_rate = 0.0
        self.link_capacity = 0.0
        self.bytes_sent = 0.0
        self.bytes_recv = 0.0
        self.bytes_sent_delta=0
        self.bytes_recv_delta=0
        self.packets_sent = 0
        self.packets_recv = 0
        self.packets_sent_delta=0
        self.packets_recv_delta=0
        self.dropin = 0
        self.dropout = 0
        self.dropin_delta=0
        self.dropout_delta=0
        self.nic_speed = 0  # this is in mb=megabits
        self.nic_mtu = 0  # max transmission speed of nic
        # identifying properties
        self.start_time = ""
        self.end_time = ""
        self.count = 0
        self.latency = []


    def set_user(self, user_passed):
        user = os.getenv('ODS_USER', '')
        if len(user_passed) > 0:
            self.ods_user = user_passed
        else:
            self.ods_user = user

    def measure(self, interface='', measure_tcp=True, measure_udp=True, measure_kernel=True, measure_network=True, print_to_std_out=False, latency_host="http://google.com"):
        self.start_time = datetime.now().__str__()
        self.interface = interface
        if measure_kernel:
            self.measure_kernel()
        if measure_network:
            print('Getting metrics of: ' + interface)
            # we could take the average of all speeds that every socket experiences and thus get a rough estimate of bandwidth??
            self.measure_network(interface)
            self.measure_latency_rtt(latency_host)
        if measure_tcp:
            print('Measuring tcp')
            psutil.net_connections(kind="tcp")
        if measure_udp:
            print('Measuring udp')
            psutil.net_connections(kind="udp")
        self.end_time = datetime.now().__str__()
        if(print_to_std_out):
            print("\n", json.dumps(self.__dict__), "\n")
        # self.to_file()

    def measure_kernel(self):
        self.active_core_count = multiprocessing.cpu_count()
        if platform.system() != 'Darwin':
            freq = psutil.cpu_freq()
            self.cpu_frequency_max = freq[2]
            self.cpu_frequency_current = freq[0]
            self.cpu_frequency_min = freq[1]
            print(freq)
        self.cpu_arch = platform.platform()
        self.active_core_count = multiprocessing.cpu_count()

    def measure_latency_rtt(self, latency_host="http://google.com"):
        self.latency = statistics.mean(measure_latency(host="google.com")) #in miliseconds and a list
        r=requests.get(latency_host)
        self.rtt = r.elapsed.microseconds/1000
        print('Latency =', self.latency, " RTT=", self.rtt)
        print('LatencyType =', type(self.latency), " RTT Type=", type(self.rtt))

    def measure_network(self, interface):
        nic_counter_dic = psutil.net_io_counters(pernic=True, nowrap=True)
        if interface not in nic_counter_dic:
            raise Exception("The interface passed was not found on your system")
        interface_counter_tuple = nic_counter_dic[interface]
        self.bytes_sent = interface_counter_tuple[0]
        self.bytes_recv = interface_counter_tuple[1]
        self.packets_sent = interface_counter_tuple[2]
        self.packets_recv = interface_counter_tuple[3]
        self.errin = interface_counter_tuple[4]
        self.errout = interface_counter_tuple[5]
        self.dropin = interface_counter_tuple[6]
        self.dropout = interface_counter_tuple[7]
        sys_interfaces = psutil.net_if_stats()
        interface_stats = sys_interfaces[self.interface]
        self.nic_mtu = interface_stats[3]
        self.nic_speed = interface_stats[2]

    def get_system_interfaces(self):
        nic_counter_dic = psutil.net_io_counters(pernic=True, nowrap=True)
        return nic_counter_dic.keys()

    def to_file(self, folder_path, folder_name=".pmeter", file_name="pmeter_measure.txt"):
        if folder_path is None:
            folder_path = os.path.join(Path.home(), ".config", folder_name)

        folder_path = os.path.join(folder_path, folder_name)
        full_path = os.path.join(folder_path, file_name)
        print(full_path)
        os.makedirs(os.path.dirname(full_path), exist_ok=True)
        j = json.dumps(self.__dict__)
        with open(full_path, "a+") as f:
            f.write(j + "\n")

    def do_deltas(self, old_metric):
        self.bytes_sent_delta = self.bytes_sent - old_metric.bytes_sent
        self.bytes_recv_delta = self.bytes_recv - old_metric.bytes_recv
        self.packets_sent_delta = self.packets_sent - old_metric.packets_sent
        self.packets_recv_delta = self.packets_recv - old_metric.packets_recv
        self.errin_delta = self.errin - old_metric.errin
        self.errout_delta = self.errout - old_metric.errout
        self.dropin_delta = self.dropin - old_metric.dropin
        self.dropout_delta = self.dropout - old_metric.dropout





import setuptools,os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open('requirements.txt') as f:
    requirements = f.readlines()
with open("_version.py",encoding='utf-8') as f:
    version = f.read().strip().split(" = ")[1][1:-1]

setuptools.setup(
    name="pmeter_ods",
    version=version,
    author="OneDataShare",
    author_email="onedatashare@gmail.com",
    description="A CLI tool that helps capture metrics from the Operating System",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/didclab/pmeter",
    project_urls={},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(exclude=['tests']),
    entry_points ={
            'console_scripts': [
                'pmeter = pmeter.pmeter_cli:main'
            ]
        },
    install_requires = [requirements],
    python_requires=">=3.7",
)
